/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package com.coyotesong.compress.archivers.zip;

import java.util.Random;
import java.util.zip.CRC32;

/**
 * Traditional PKWARE Encryption.  See APPNOTE_6.2.0.txt at line 1667 et seq.
 * 
 * Important note from the documentation: <em>This
 * form of encryption is considered weak by today's standards
 * and its use is recommended only for situations with
 * low security needs or for compatibility with older .ZIP applications.</em>
 * In practice this means that it will take someone less than a second to
 * crack the encryption with the right software.
 *
 * Technical note: this is a conventional feedback stream cipher and ANY function
 * <code>f(void)</code> will work as long as it's consistent. This means that we
 * can't simply verify an encryption/decryption cycle - we need to verify that we
 * can encrypt and decrypt content provided by another implementation of zip.
 *  
 * @author Bear Giles <bgiles@coyotesong.com>
 */
public class PkwareEncryptionAware {
	private CRC32 crc32 = new CRC32();
    private int[] key;
    
    // we normally want a strong PRNG but these values will be public salt.
    private Random random = new Random();

    /**
     * Set the encryption/decryption password.
     * 
     * @param password
     */
    public void setPassword(String password) {
        key = new int[3];
        key[0] = 305419896;
        key[1] = 591751049;
        key[2] = 878082192;

        byte[] bytes = password.getBytes();
        for (int i = 0; i < bytes.length; i++) {
            updateKeys(bytes[i]);
        }    
    }
    
    /**
     * Convenience method to calculate CRC value.
     * 
     * @param key
     * @param b
     * @return
     */
    private int crc32(int key, byte b) {
    	crc32.reset();
    	crc32.update(key);
    	crc32.update(b);
    	return (int) crc32.getValue();
    }

    /**
     * Update the keys
     * @param b
     */
    private void updateKeys(byte b) {
    	key[0] = crc32(key[0], b);
        key[1] = (key[1] + (key[0] & 0xFF)) * 134775813 + 1;
        key[2] = crc32(key[2], (byte) ((key[1] >> 24) & 0xFF));
    }

    /**
     * Initialize salt. Note the final 1 or two bytes
     * will be the high-order word/byte of the CRC for the
     * file being decrypted. This can be used as a quick
     * check whether the provided password is valid.
     */
    public void initialize(byte[] header) {
        decryptBlock(header, 0, header, 0, 12);
    }

    /**
     * Get pseudorandom value.
     * 
     * @return
     */
    public byte f() {
        byte t = (byte) ((key[2] | 2) & 0xFF);
        return (byte) ((t * (t ^ 1)) >> 8);
    }

    /**
     * Encrypt a full message including salt. This uses the 'new' format
     * with a single byte of CRC data. The 'old' format (prior to version 2.0)
     * used the top 2 bytes of the CRC.
     */
    public byte[] encrypt(byte[] src) {
    	return encrypt(src, 0, src.length);
    }

    /**
     * Encrypt a full message including salt. This uses the 'new' format
     * with a single byte of CRC data. The 'old' format (prior to version 2.0)
     * used the top 2 bytes of the CRC.
     */
    public byte[] encrypt(byte[] src, int off, int len) {
    	byte[] salt = new byte[12];
    	byte[] dest = new byte[len + 12];

    	// initialize salt, including checksum
    	random.nextBytes(salt);
    	crc32.reset();
    	crc32.update(src, off, len);
    	salt[11] = (byte) ((crc32.getValue() >> 24) & 0xFF); // overrides last byte
    	System.arraycopy(salt, 0, dest, 0, 12);
    	
    	// encrypt message
    	initialize(salt);
    	encryptBlock(src, off, dest, 12, len);
 
    	return dest;
    }
       
    /**
     * Encrypt a block without the header.
     */
    public void encryptBlock(byte[] src, int srcoff, byte[] dest, int destoff, int len) {
        for (int i = 0; i < len; i++) {
           byte t = src[i + srcoff];
           dest[i + destoff] = (byte) (t ^ f());
           updateKeys(t);
        }
    }

    /**
     * Decrypt a full message including salt. This method returns NULL if the password
     * was invalid. (There is still a 1-in-256 chance that we'll get a buffer containing garbage.)
     */
    public byte[] decrypt(byte[] src) {
    	return decrypt(src, 0, src.length);
    }

    /**
     * Decrypt a full message including salt. This method returns NULL if the password
     * was invalid. (There is still a 1-in-256 chance that we'll get a buffer containing garbage.)
     */
    public byte[] decrypt(byte[] src, int off, int len) {
    	if (len < 12) {
    		throw new IllegalArgumentException("length must be at least 12 bytes");
    	}

    	byte[] dest = new byte[len - 12];
    	
    	// decrypt message
    	byte[] salt = new byte[12];
    	System.arraycopy(src, off, salt, 0, 12);
    	initialize(salt);
    	decryptBlock(src, off + 12, dest, 0, len - 12);

    	// verify checksum
    	crc32.reset();
    	crc32.update(dest);
    	if (src[off + 11] != (byte) ((crc32.getValue() >> 24) & 0xFF)) {
    		return null;
    	}

    	return dest;
    }
    
    /**
     * Decrypt a buffer
     */
    public void decryptBlock(byte[] src, int srcoff, byte[] dest, int destoff, int len) {
        for (int i = 0; i < len; i++) {
           byte t = (byte) (src[i + srcoff] ^ f());
           updateKeys(t);
           dest[i + destoff] = t;
        }
    }
    
    // quick test.
    public static void main(String[] args) {
    	String password = "password";

    	PkwareEncryptionAware p = new PkwareEncryptionAware();
    	p.setPassword(password);
    	byte[] ctext = p.encrypt("foobar".getBytes());

    	p = new PkwareEncryptionAware();
    	p.setPassword(password);
    	byte[] ptext = p.decrypt(ctext);
    	
    	System.out.printf("text: '%s'\n", new String(ptext));
    }
}
